import asyncio
import enum


class ASGICycleState(enum.Enum):
    REQUEST = enum.auto()
    RESPONSE = enum.auto()


class ASGICycle:
    def __init__(self, scope: dict, body: bytes = b"") -> None:
        self.scope = scope
        self.body = body
        self.state = ASGICycleState.REQUEST
        self.app_queue = None
        self.response = {}

    def __call__(self, app) -> dict:
        loop = asyncio.new_event_loop()

        self.app_queue = asyncio.Queue(loop=loop)
        self.put_message({"type": "http.request", "body": self.body})

        asgi_instance = app(self.scope)
        asgi_task = loop.create_task(asgi_instance(self.receive, self.send))
        loop.run_until_complete(asgi_task)

        return self.response

    def put_message(self, message: dict) -> None:
        self.app_queue.put_nowait(message)

    async def receive(self) -> dict:
        message = await self.app_queue.get()
        return message

    async def send(self, message: dict) -> None:
        message_type = message["type"]

        if self.state is ASGICycleState.REQUEST:
            if message_type != "http.response.start":
                raise RuntimeError(
                    f"Expected 'http.response.start', received: {message_type}"
                )

            status_code = message["status"]
            headers = {
                k.decode("utf-8"): v.decode("utf-8")
                for k, v in message.get("headers", [])
            }

            self.on_response_start(headers, status_code)
            self.state = ASGICycleState.RESPONSE

        elif self.state is ASGICycleState.RESPONSE:
            if message_type != "http.response.body":
                raise RuntimeError(
                    f"Expected 'http.response.body', received: {message_type}"
                )

            body = message["body"]
            if isinstance(body, bytes):
                body = body.decode("utf-8")

            self.on_response_body(body)
            self.put_message({"type": "http.disconnect"})

    def on_response_start(self, headers: list, status_code: int) -> None:
        raise NotImplementedError()  # pragma: no cover

    def on_response_body(self, body: str) -> None:
        raise NotImplementedError()  # pragma: no cover



import urllib.parse
import cgi
from azure.functions import HttpResponse
from mangum.handlers.asgi import ASGICycle
from mangum.utils import encode_query_string


class AzureFunctionCycle(ASGICycle):
    def on_response_start(self, headers: dict, status_code: int) -> None:
        self.response["status_code"] = status_code
        self.response["headers"] = headers
        mimetype = None
        charset = None
        if "content-type" in headers:
            mimetype, options = cgi.parse_header(headers["content-type"])
            charset = options.get("charset", None)
        self.response["mimetype"] = mimetype
        self.response["charset"] = charset

    def on_response_body(self, body: bytes) -> None:
        self.response["body"] = body


def azure_handler(app, req) -> dict:

    server = None
    client = None
    scheme = "https"
    method = req.method
    headers = req.headers.items()
    parsed = urllib.parse.urlparse(req.url)
    scheme = parsed.scheme
    path = parsed.path
    query_string = encode_query_string(req.params) if req.params else ""

    scope = {
        "type": "http",
        "server": server,
        "client": client,
        "method": method,
        "path": path,
        "scheme": scheme,
        "http_version": "1.1",
        "root_path": "",
        "query_string": query_string,
        "headers": headers,
    }

    body = req.get_body() or b""
    if not isinstance(body, bytes):
        body = body.encode("utf-8")

    asgi_cycle = AzureFunctionCycle(scope, body=body)
    response = asgi_cycle(app)

    return HttpResponse(
        body=response["body"],
        headers=response["headers"],
        status_code=response["status_code"],
        mimetype=response["mimetype"],
        charset=response["charset"],
    )



import base64
from mangum.handlers.asgi import ASGICycle
from mangum.utils import encode_query_string


class AWSLambdaCycle(ASGICycle):
    def on_response_start(self, headers: dict, status_code: int) -> None:
        self.response["statusCode"] = status_code
        self.response["isBase64Encoded"] = False
        self.response["headers"] = headers

    def on_response_body(self, body: str) -> None:
        self.response["body"] = body


def aws_handler(app, event: dict, context: dict) -> dict:
    server = None
    client = None
    method = event["httpMethod"]
    headers = event["headers"] or {}
    path = event["path"]
    host = headers.get("Host")
    scheme = headers.get("X-Forwarded-Proto", "http")
    x_forwarded_for = headers.get("X-Forwarded-For")
    x_forwarded_port = headers.get("X-Forwarded-Port")

    if x_forwarded_port and x_forwarded_for:
        port = int(x_forwarded_port)
        client = (x_forwarded_for, port)
        if host:
            server = (host, port)

    query_string_params = event["queryStringParameters"]
    query_string = (
        encode_query_string(query_string_params) if query_string_params else ""
    )

    scope = {
        "server": server,
        "client": client,
        "scheme": scheme,
        "root_path": "",
        "query_string": query_string,
        "headers": headers.items(),
        "type": "http",
        "http_version": "1.1",
        "method": method,
        "path": path,
    }

    body = event["body"] or b""
    if body and event.get("isBase64Encoded"):
        body = base64.standard_b64decode(body)

    handler = AWSLambdaCycle(scope, body=body)
    return handler(app)
